from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Type

import equinox as eqx
import jax.numpy as jnp
from equinox import filter_jit
from jax import Array, vmap
from jax.lax import cond

from .engines import ComputationEngine, DenseEngine
from .utils import format_jax_array
from .module import AbstractModule

if TYPE_CHECKING:
	pass


class AbstractKernel(AbstractModule):
	static_class: ClassVar[Optional[Type[StaticAbstractKernel]]] = None
	computation_engine: Type[ComputationEngine] = eqx.field(
		static=True, default=DenseEngine, kw_only=True
	)

	def __init__(self, computation_engine=DenseEngine, **kwargs):
		"""
		Initialize the kernel.

		Args:
			computation_engine: The computation engine to use for covariance calculations
			**kwargs: Additional keyword arguments (for subclass compatibility)
		"""
		super().__init__(**kwargs)

		# Set the computation engine
		self.computation_engine = computation_engine


	@filter_jit
	def __call__(self, x1: Array, x2: Optional[Array] = None) -> Array:
		# If no x2 is provided, we compute the covariance between x1 and itself
		if x2 is None:
			x2 = x1

		# Turn scalar inputs into vectors
		x1, x2 = jnp.atleast_1d(x1), jnp.atleast_1d(x2)

		# Ensure static_class is not None
		assert self.static_class is not None, "static_class must be defined in subclass"

		# Call the appropriate method
		if jnp.ndim(x1) == 1 and jnp.ndim(x2) == 1:
			return self.computation_engine.pairwise_cov_if_not_nan(self, x1, x2)
		elif jnp.ndim(x1) == 2 and jnp.ndim(x2) == 1:
			return self.computation_engine.cross_cov_vector_if_not_nan(self, x1, x2)
		elif jnp.ndim(x1) == 1 and jnp.ndim(x2) == 2:
			return self.computation_engine.cross_cov_vector_if_not_nan(self, x2, x1)
		elif jnp.ndim(x1) == 2 and jnp.ndim(x2) == 2:
			return self.computation_engine.cross_cov_matrix(self, x1, x2)
		else:
			raise ValueError(
				f"Invalid input dimensions: x1 has shape {x1.shape}, x2 has shape {x2.shape}. "
				"Expected scalar, 1D or 2D arrays as inputs."
			)


class StaticAbstractKernel:
	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute the kernel covariance value between two vectors.

		:param kern: kernel instance containing the hyperparameters
		:param x1: scalar array
		:param x2: scalar array
		:return: scalar array
		"""
		return jnp.array(jnp.nan)  # To be overwritten in subclasses

	@classmethod
	@filter_jit
	def pairwise_cov_if_not_nan(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Returns NaN if either x1 or x2 is NaN, otherwise calls the pairwise_cov method.

		:param kern: kernel instance containing the hyperparameters
		:param x1: scalar array
		:param x2: scalar array
		:return: scalar array
		"""
		return cond(  # type: ignore[no-any-return]
			jnp.any(jnp.isnan(x1) | jnp.isnan(x2)),
			lambda _: jnp.nan,
			lambda _: cls.pairwise_cov(kern, x1, x2),
			None,
		)

	@classmethod
	@filter_jit
	def cross_cov_vector(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute the kernel cross covariance values between an array of vectors (matrix) and a vector.

		:param kern: kernel instance containing the hyperparameters
		:param x1: vector array (N, )
		:param x2: scalar array
		:return: vector array (N, )
		"""
		return vmap(lambda x: cls.pairwise_cov_if_not_nan(kern, x, x2), in_axes=0)(x1)

	@classmethod
	@filter_jit
	def cross_cov_vector_if_not_nan(
		cls, kern: AbstractKernel, x1: Array, x2: Array, **kwargs
	) -> Array:
		"""
		Returns an array of NaN if scalar is NaN, otherwise calls the cross_cov_vector method.

		:param kern: kernel instance containing the hyperparameters
		:param x1: vector array (N, )
		:param x2: scalar array
		:param kwargs: hyperparameters of the kernel
		:return: vector array (N, )
		"""
		return cond(  # type: ignore[no-any-return]
			jnp.any(jnp.isnan(x2)),
			lambda _: jnp.full(len(x1), jnp.nan),
			lambda _: cls.cross_cov_vector(kern, x1, x2),
			None,
		)

	@classmethod
	@filter_jit
	def cross_cov_matrix(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute the kernel covariance matrix between two vector arrays.

		:param kern: kernel instance containing the hyperparameters
		:param x1: vector array (N, )
		:param x2: vector array (M, )
		:return: matrix array (N, M)
		"""
		return vmap(lambda x: cls.cross_cov_vector_if_not_nan(kern, x2, x), in_axes=0)(x1)



from __future__ import annotations

import equinox as eqx

from ..module import AbstractModule
from ..other.ConstantKernel import ConstantKernel


class WrapperModule(AbstractModule):
	"""Base class for modules that wrap another module and transform its output."""

	inner: AbstractModule = eqx.field()

	def __init__(self, inner=None, **kwargs):
		if not isinstance(inner, AbstractModule):
			inner = ConstantKernel(value=inner)

		super().__init__(**kwargs)
		self.inner = inner

	def replace(self, **kwargs):
		wrapper_kwargs = {}
		inner_kwargs = {}

		for k, v in kwargs.items():
			if k == "inner" or (hasattr(self, k) and k != "inner" and hasattr(type(self), k)):
				wrapper_kwargs[k] = v
			else:
				inner_kwargs[k] = v

		result = self
		if wrapper_kwargs:
			result = super().replace(**wrapper_kwargs)

		if inner_kwargs:
			new_inner = result.inner.replace(**inner_kwargs)
			result = eqx.tree_at(lambda s: s.inner, result, new_inner)

		return result


import equinox as eqx
from equinox import filter_jit
from jax import Array
from jax import numpy as jnp

from ..AbstractKernel import AbstractKernel
from .DotProductKernel import StaticDotProductKernel


class StaticLinearKernel(StaticDotProductKernel):
	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute the linear kernel covariance value between two vectors.

		:param kern: the kernel to use, containing hyperparameters (variance_b, variance_v, offset_c).
		:param x1: scalar array.
		:param x2: scalar array.
		:return: scalar array (covariance value).
		"""
		kern = eqx.combine(kern)
		x1_shifted = x1 - kern.offset_c  # type: ignore[attr-defined]
		x2_shifted = x2 - kern.offset_c  # type: ignore[attr-defined]

		# Compute the dot product of the shifted vectors
		dp = cls.distance_func(x1_shifted, x2_shifted)

		return kern.variance_b + kern.variance_v * dp  # type: ignore[no-any-return,attr-defined]


class LinearKernel(AbstractKernel):
	"""
	Linear Kernel

	Parameters variance_b and variance_v must be non-negative (>= 0).
	Parameter offset_c can be any real value.

	Note: Variances are NOT transformed (no log-exp trick) because they can be zero,
	which is incompatible with log-based transformations.
	"""

	variance_b: Array = eqx.field(converter=jnp.asarray)
	variance_v: Array = eqx.field(converter=jnp.asarray)
	offset_c: Array = eqx.field(converter=jnp.asarray)
	static_class = StaticLinearKernel

	def __init__(self, variance_b, variance_v, offset_c, **kwargs):
		"""
		Initialize the Linear kernel.

		Args:
			variance_b: Bias variance (σ²_b, must be non-negative). Controls the vertical offset.
			variance_v: Weight variance (σ²_v, must be non-negative). Controls the slope.
			offset_c: Input offset (c, can be any real value). Determines the crossing point.

		Raises:
			ValueError: If variance_b or variance_v is negative
		"""
		# Validate non-negativity for variances
		variance_b = jnp.array(variance_b)
		variance_v = jnp.array(variance_v)

		variance_b = eqx.error_if(
			variance_b, jnp.any(variance_b < 0), "variance_b must be non-negative."
		)
		variance_v = eqx.error_if(
			variance_v, jnp.any(variance_v < 0), "variance_v must be non-negative."
		)

		# Initialize parent (locks config)
		super().__init__(**kwargs)

		# Store parameters as-is (no transformation)
		# Variances can be 0, which is incompatible with log-based transforms
		self.variance_b = jnp.asarray(variance_b)
		self.variance_v = jnp.asarray(variance_v)
		self.offset_c = jnp.asarray(offset_c)



import equinox as eqx
from equinox import filter_jit
from jax import Array
from jax import numpy as jnp

from ..AbstractKernel import AbstractKernel
from ..distances import squared_euclidean_distance
from .StationaryKernel import StaticStationaryKernel


class StaticRationalQuadraticKernel(StaticStationaryKernel):
	distance_func = squared_euclidean_distance

	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute the Rational Quadratic kernel covariance value between two vectors.

		:param kern: the kernel to use, containing hyperparameters (length_scale, alpha)
		:param x1: scalar array
		:param x2: scalar array
		:return: covariance value (scalar)
		"""
		kern = eqx.combine(kern)
		squared_dist = cls.distance_func(x1, x2)

		base = 1 + squared_dist / (2 * kern.alpha * kern.length_scale**2)  # type: ignore[attr-defined]

		return jnp.power(base, -kern.alpha)  # type: ignore[attr-defined]


class RationalQuadraticKernel(AbstractKernel):
	"""
	Rational Quadratic Kernel

	All parameters (length_scale, alpha) are always positive.
	Internally, they may be stored in unconstrained space.
	"""

	_raw_length_scale: Array = eqx.field(converter=jnp.asarray)
	_raw_alpha: Array = eqx.field(converter=jnp.asarray)
	static_class = StaticRationalQuadraticKernel

	def __init__(self, length_scale, alpha, **kwargs):
		"""
		Initialize the Rational Quadratic kernel.

		Args:
			length_scale: length scale parameter (ℓ, must be positive)
			alpha: relative weighting of large-scale and small-scale variations (α, must be positive)

		Raises:
			ValueError: If any parameter is not positive
		"""
		# Validate positivity
		length_scale = jnp.array(length_scale)
		alpha = jnp.array(alpha)

		length_scale = eqx.error_if(
			length_scale, jnp.any(length_scale <= 0), "length_scale must be positive."
		)
		alpha = eqx.error_if(alpha, jnp.any(alpha <= 0), "alpha must be positive.")

		# Initialize parent (locks config)
		super().__init__(**kwargs)

		# Transform to unconstrained space
		from ..transforms import to_unconstrained

		self._raw_length_scale = to_unconstrained(jnp.asarray(length_scale))
		self._raw_alpha = to_unconstrained(jnp.asarray(alpha))

	@property
	def length_scale(self) -> Array:
		"""Get the length scale in constrained space (always positive)."""
		from ..transforms import to_constrained

		return to_constrained(self._raw_length_scale)

	@property
	def alpha(self) -> Array:
		"""Get the alpha parameter in constrained space (always positive)."""
		from ..transforms import to_constrained

		return to_constrained(self._raw_alpha)



from __future__ import annotations

from typing import Optional

import jax.numpy as jnp
from equinox import filter_jit
from jax import Array

from .WrapperModule import WrapperModule


class ExpModule(WrapperModule):
	"""Module that applies the exponential function to the output of another module."""

	@filter_jit
	def __call__(self, x1: Array, x2: Optional[Array] = None) -> Array:
		if x2 is None:
			return jnp.exp(self.inner(x1))
		return jnp.exp(self.inner(x1, x2))

	def __str__(self):
		return f"Exp({self.inner})"


import equinox as eqx
from equinox import filter_jit
from jax import Array
from jax import numpy as jnp

from ..AbstractKernel import AbstractKernel
from ..distances import squared_euclidean_distance
from ..transforms import to_constrained, to_unconstrained
from .StationaryKernel import StaticStationaryKernel


class StaticSEKernel(StaticStationaryKernel):
	distance_func = squared_euclidean_distance

	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute the kernel covariance value between two vectors.

		:param kern: kernel instance containing the hyperparameters
		:param x1: scalar array
		:param x2: scalar array
		:return: scalar array
		"""
		kern = eqx.combine(kern)
		return jnp.exp(-0.5 * cls.distance_func(x1, x2) / kern.length_scale**2)  # type: ignore[attr-defined]


class SEKernel(AbstractKernel):
	"""
	Squared Exponential (aka "RBF" or "Gaussian") Kernel

	The length_scale parameter is always positive. Internally, it may be stored in an
	unconstrained space (log-space or softplus-inverse space) depending on the global
	configuration, which improves optimization stability.
	"""

	_raw_length_scale: Array = eqx.field(converter=jnp.asarray)
	static_class = StaticSEKernel

	def __init__(self, length_scale, **kwargs):
		"""
		Initialize the SE kernel with a length scale parameter.

		Args:
			length_scale: Length scale parameter (must be positive). This is provided
				in the constrained space (positive values) and will be converted to
				the appropriate unconstrained space based on config.parameter_transform.
			**kwargs: Additional arguments passed to AbstractKernel (e.g., computation_engine)

		Raises:
			ValueError: If length_scale is not positive
		"""
		# Assert length_scale is positive
		length_scale = jnp.array(length_scale)
		length_scale = eqx.error_if(length_scale, jnp.any(length_scale <= 0), "length_scale must be positive.")

		# Initialize parent (marks kernels as instantiated, locks config)
		super().__init__(**kwargs)

		# Transform to unconstrained space
		self._raw_length_scale = to_unconstrained(jnp.asarray(length_scale))

	@property
	def length_scale(self) -> Array:
		"""
		Get the length scale in constrained space (always positive).

		This property applies the forward transformation to convert from the internal
		unconstrained representation back to the constrained (positive) space.

		Returns:
			Length scale parameter (positive)
		"""
		return to_constrained(self._raw_length_scale)


class RBFKernel(SEKernel):
	"""
	Same as SEKernel
	"""

	pass



from __future__ import annotations

from typing import TYPE_CHECKING

import jax.numpy as jnp
from equinox import error_if, filter_jit
from jax import Array, vmap
from jax.lax import cond

if TYPE_CHECKING:
	from .AbstractKernel import AbstractKernel

class ComputationEngine:
	"""
	Superclass for all computation engines.

	A computation engine defines how to build cross-covariance vectors/matrices given inputs and a
	kernel.
	Most of the time, the `DenseEngine` is the one wanted, as full covariance matrices are computed.
	It is the default engine associated to all kernels.

	But some other engines can be used to compute only specific parts of the covariance matrices,
	like `DiagonalEngine`, which only computes the diagonal of the covariance matrix.

	Some other engines can also be used when treating inputs with specific structures, like the
	`RegularGridEngine`, which exploits the regular grid structure of the inputs to compute the
	covariance matrix more efficiently.

	For maximal efficiency, some engines might require specific constraints on the inputs or the
	kernel. Asserting these constraints at runtime is sometimes costly. This is why most engines
	come in two versions: `Safe` and `Fast`.

	The `Safe` versions either check the constraints at runtime, or adapt their computations to
	always satisfy the constraints.

	The `Fast` versions skip any check, and assume that the user ensured that the constraints
	are satisfied. Failing to do so may lead to incorrect results or runtime errors. However, you
	can still use the `check_constraints` method to verify that the inputs and kernel are compatible.

	An example of this is using a `RegularGridEngine`. This engine is more efficient when
	the inputs are indeed on a regular grid and the kernel is stationary, as the covariance matrix
	is just a repetition of the same vector, offset by some indices.
	However, checking that the inputs are on a regular grid (according to the metric used by the
	kernel) is costly. The `Safe` version of the engine will check this at runtime, while the `Fast`
	version will skip this check.

	Computation engines are an interface between the kernels and their abstract implementation.
	They should implement all the methods defined in this superclass.
	"""
	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute the kernel covariance value between two vectors.

		:param kern: kernel instance containing the hyperparameters
		:param x1: scalar array
		:param x2: scalar array
		:return: scalar array
		"""
		return kern.static_class.pairwise_cov(kern, x1, x2)  # type: ignore[union-attr]

	@classmethod
	@filter_jit
	def pairwise_cov_if_not_nan(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Returns NaN if either x1 or x2 is NaN, otherwise calls the pairwise_cov method.

		:param kern: kernel instance containing the hyperparameters
		:param x1: scalar array
		:param x2: scalar array
		:return: scalar array
		"""
		return kern.static_class.pairwise_cov_if_not_nan(kern, x1, x2)  # type: ignore[union-attr]

	@classmethod
	@filter_jit
	def cross_cov_vector(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute the kernel cross covariance values between an array of vectors (matrix) and a vector.

		:param kern: kernel instance containing the hyperparameters
		:param x1: vector array (N, )
		:param x2: scalar array
		:return: vector array (N, )
		"""
		return kern.static_class.cross_cov_vector(kern, x1, x2)  # type: ignore[union-attr]

	@classmethod
	@filter_jit
	def cross_cov_vector_if_not_nan(
		cls, kern: AbstractKernel, x1: Array, x2: Array, **kwargs
	) -> Array:
		"""
		Returns an array of NaN if scalar is NaN, otherwise calls the cross_cov_vector method.

		:param kern: kernel instance containing the hyperparameters
		:param x1: vector array (N, )
		:param x2: scalar array
		:param kwargs: hyperparameters of the kernel
		:return: vector array (N, )
		"""
		return kern.static_class.cross_cov_vector_if_not_nan(kern, x1, x2, **kwargs)  # type: ignore[union-attr]

	@classmethod
	@filter_jit
	def cross_cov_matrix(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute the kernel covariance matrix between two vector arrays.

		:param kern: kernel instance containing the hyperparameters
		:param x1: vector array (N, )
		:param x2: vector array (M, )
		:return: matrix array (N, M)
		"""
		return kern.static_class.cross_cov_matrix(kern, x1, x2)  # type: ignore[union-attr]

	@classmethod
	@filter_jit
	def check_constraints(cls, kern: AbstractKernel, x1: Array, x2: Array) -> bool:
		"""
		Check if the inputs and kernel are compatible with the computation engine.

		By default, all inputs and kernels are compatible. Specific engines can override this method
		to implement specific checks, using equinox's `error_if` fonction.

		:param kern: kernel instance containing the hyperparameters
		:param x1: vector array (N, )
		:param x2: vector array (M, )
		:return: True if the inputs are compatible, False otherwise
		"""
		return True


class DenseEngine(ComputationEngine):
	"""
	Dense computation engine.

	Computes full covariance matrices between inputs.
	"""
	pass  # Super-class already implements all needed methods


class FastDiagonalEngine(ComputationEngine):
	"""
	Fast Diagonal computation engine. Returns covariance value if inputs are equal, otherwise 0.
	This engine assumes that the two sets of inputs are identical (meaning that for every i, x1[i] == x2[i]).
	This results in a diagonal covariance matrix.

	An additional speedup is possible when computing it on a stationary of constant kernel, as the
	covariance values are all identical. In this case, the diagonal is just a repetition of the same value.
	"""
	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern, x1: Array, x2: Array) -> Array:
		# It is assumed that x1 == x2
		return kern.static_class.pairwise_cov(kern, x1, x2)  # type: ignore[no-any-return]

	@classmethod
	@filter_jit
	def cross_cov_matrix(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute only the diagonal of the kernel covariance matrix between two identical vector arrays.
		It is assumed that for every i, x1[i] == x2[i].

		:param kern: kernel instance containing the hyperparameters
		:param x1: vector array (N, )
		:param x2: vector array (M, )
		:return: vector array (min(N, M), ) containing the diagonal elements
		"""
		# Import here to avoid circular imports
		from .other.ConstantKernel import ConstantKernel
		from .stationary.StationaryKernel import StaticStationaryKernel

		# Check whether kern inherits from StationaryKernel or ConstantKernel for speedup
		if isinstance(kern.static_class, StaticStationaryKernel) or isinstance(kern, ConstantKernel):
			# Single covariance value for all diagonal elements
			return jnp.eye(x1.shape[0]) * cls.pairwise_cov_if_not_nan(kern, x1[0], x2[0])

		# Else, compute all diagonal elements
		return jnp.eye(x1.shape[0]) * vmap(cls.pairwise_cov_if_not_nan, in_axes=(None, 0, 0))(kern, x1, x2)

	@classmethod
	@filter_jit
	def check_constraints(cls, kern: AbstractKernel, x1: Array, x2: Array) -> bool:
		"""
		Check that the two sets of inputs are identical.

		:param kern: kernel instance containing the hyperparameters
		:param x1: vector array (N, )
		:param x2: vector array (M, )
		:return: True if the inputs are compatible, False otherwise
		"""
		# Check that the two sets of inputs are identical
		x1 = error_if(x1, jnp.any(x1 != x2), "Inputs in x1 and x2 do not match")
		return True


class SafeDiagonalEngine(ComputationEngine):
	"""
	Diagonal computation engine. Returns covariance value if inputs are equal, otherwise 0.
	This engine doesn't require any constraint on inputs or kernel, but may be inefficient
	if you always compute the cross-covariance between two equivalent sets of inputs. In this case,
	use FastDiagonalEngine instead.

	When computing on the same set of inputs, this results in a diagonal covariance matrix.
	"""
	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern, x1: Array, x2: Array) -> Array:
		return cond(  # type: ignore[no-any-return]
			jnp.all(x1 == x2), lambda _: kern.static_class.pairwise_cov(kern, x1, x2), lambda _: jnp.array(0.0), None
		)

	@classmethod
	@filter_jit
	def cross_cov_matrix(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute diagonal covariance matrix with conditional checks for input equality.
		Returns kernel value when x1[i] == x2[i], otherwise 0.

		:param kern: kernel instance containing the hyperparameters
		:param x1: vector array (N, )
		:param x2: vector array (M, )
		:return: diagonal matrix array (N, N) where off-diagonal elements are 0
		"""
		# Import here to avoid circular imports
		from .other.ConstantKernel import ConstantKernel
		from .stationary.StationaryKernel import StaticStationaryKernel

		# For efficiency with same inputs, check if all inputs are identical
		all_same = jnp.all(x1 == x2)

		def compute_full_diagonal():
			# Compute diagonal elements where x1[i] == x2[i], 0 otherwise
			return jnp.eye(x1.shape[0]) * vmap(cls.pairwise_cov, in_axes=(None, 0, 0))(kern, x1, x2)

		def compute_fast_diagonal():
			# All inputs are same, can optimize
			if isinstance(kern.static_class, StaticStationaryKernel) or isinstance(kern, ConstantKernel):
				# Single covariance value for all diagonal elements
				return jnp.eye(x1.shape[0]) * kern.static_class.pairwise_cov(kern, x1[0], x2[0])
			else:
				# Compute all diagonal elements
				return jnp.eye(x1.shape[0]) * vmap(kern.static_class.pairwise_cov, in_axes=(None, 0, 0))(kern, x1, x2)  # type: ignore[union-attr]

		return cond(all_same, lambda _: compute_fast_diagonal(), lambda _: compute_full_diagonal(), None)  # type: ignore[no-any-return]


class FastRegularGridEngine(ComputationEngine):
	"""
	Fast Regular Grid computation engine.

	Exploits the regular grid structure of the inputs to compute the covariance matrix more efficiently.
	This engine assumes that the inputs are indeed on a regular grid according to the metric
	used by the kernel. Failing to do so may lead to incorrect results or runtime errors.
	However, you can still use the `check_constraints` method to verify that the inputs and kernel
	are compatible.
	"""
	@classmethod
	@filter_jit
	def cross_cov_matrix(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		# We create a circulant matrix from the first row
		n = x1.shape[0]
		g = jnp.arange(n)
		vec = kern.static_class.cross_cov_vector_if_not_nan(kern, x1, x2[0])  # type: ignore[union-attr]
		return vmap(lambda i: vec[(g + i)%n])(g)


	@classmethod
	@filter_jit
	def check_constraints(cls, kern: AbstractKernel, x1: Array, x2: Array):
		"""
		Check that the inputs are on a regular grid according to the metric used by the kernel.

		Note: we cannot only check that the kernel is stationary, as composition of stationary kernels
		can also be stationary. We perform this check by computing the k+1 diagonal of the covariance matrix
		and verifying that all values are identical.

		:param kern: kernel instance containing the hyperparameters
		:param x1: vector array (N, )
		:param x2: vector array (M, )
		:return: True if the inputs are compatible, False otherwise
		"""
		# Check that the inputs are on a regular grid according to kern.distance_func, meaning that
		# the distances between consecutive points are all identical
		diag_covs = vmap(kern.static_class.pairwise_cov, in_axes=(None, 0, 0))(kern, x1[:-1], x2[1:])  # type: ignore[union-attr]
		error_if(diag_covs, jnp.any(diag_covs != diag_covs[0]),
		         "Inputs are not on a regular grid or the kernel is not stationary.")
		return True


class SafeRegularGridEngine(FastRegularGridEngine):
	"""
	Safe Regular Grid computation engine.

	Exploits the regular grid structure of the inputs to compute the covariance matrix more efficiently.
	This engine checks at runtime that the inputs are indeed on a regular grid according to the metric
	used by the kernel.

	Note: at this time, no significant slowdown is observed when using this safe version compared to the fast version.
	"""
	@classmethod
	@filter_jit
	def cross_cov_matrix(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		# Verify constraints
		cls.check_constraints(kern, x1, x2)

		# Call parent method
		return super().cross_cov_matrix(kern, x1, x2)



import equinox as eqx
import jax.numpy as jnp
import jax.tree_util as jtu
from equinox import filter_jit
from jax import Array, vmap

from .WrapperModule import WrapperModule


class BatchModule(WrapperModule):
	"""
	Wrapper module to add batch handling to any module.

	A basic kernel usually works on inputs of shape (N, I), and produces covariance matrices of shape (N, N).

	Wrapped inside a batch module, they can either:
	- still work on inputs of shape (N, I), but produce covariance matrices of shape (B, N, N), where B is the batch size. This is useful when the hyperparameters are batched, i.e. each batch element has its own set of hyperparameters.
	- or work on inputs of shape (B, N, I), producing covariance matrices of shape (B, N, N). This is useful when the inputs are batched, regardless of whether the hyperparameters are batched or not.

	A batch module can itself be wrapped inside another batch module, to handle multiple batch dimensions/hyperparameter sets.

	This class uses vmap to vectorize the module computation over the batch dimension.
	"""
	batch_size: int = eqx.field(static=True)
	batch_in_axes: bool = eqx.field(static=True)
	batch_over_inputs: int | None = eqx.field(static=True)

	def __init__(self, inner, batch_size, batch_in_axes=None, batch_over_inputs=True, **kwargs):
		"""
		:param inner: the kernel to wrap, must be an instance of AbstractKernel
		:param batch_size: the size of the batch (int)
		:param batch_in_axes: a value or pytree indicating which hyperparameters are batched (0)
											   or shared (None) across the batch.
											   If None, all hyperparameters are assumed to be shared across the batch.
											   If 0, all hyperparameters are assumed to be batched across the batch.
											   If a pytree, it must have the same structure as inner, with hyperparameter
											   leaves being either 0 (batched) or None (shared).
		:param batch_over_inputs: whether to expect inputs of shape (B, N, I) (True) or (N, I) (False)
		"""
		# Initialize the WrapperKernel
		super().__init__(inner=inner, **kwargs)
		self.batch_size = batch_size

		# Default: all array hyperparameters are shared (None for all array leaves)
		if batch_in_axes is None:
			# Extract only array leaves and map them to None
			self.batch_in_axes = jtu.tree_map(lambda _: None, inner)
		elif batch_in_axes == 0:
			# All hyperparameters are batched
			self.batch_in_axes = jtu.tree_map(lambda _: 0, inner)
		else:
			self.batch_in_axes = batch_in_axes

		self.batch_over_inputs = 0 if batch_over_inputs else None

		# Add batch dimension to parameters where batch_in_axes is 0
		self.inner = jtu.tree_map(
			lambda param, batch_in_ax: (
				param if batch_in_ax is None else jnp.repeat(param[None, ...], batch_size, axis=0)
			),
			self.inner,
			self.batch_in_axes,
		)

	@filter_jit
	def __call__(self, x1: Array, x2: None | Array = None) -> Array:
		"""
		Compute the kernel over batched inputs using vmap.

		Args:
				x1: Input of shape (B, ..., N, I)
				x2: Optional second input of shape (B, ..., M, I)

		Returns:
				Kernel matrix of appropriate shape with batch dimension
		"""
		# Check if we can use vmap (at least one axis is not None)
		can_use_vmap = (
			not jtu.tree_all(jtu.tree_map(lambda k: k is None, self.batch_in_axes))
			or self.batch_over_inputs is not None
		)

		if can_use_vmap:
			# Use vmap when we have batched hyperparameters or batched inputs
			return vmap(  # type: ignore[no-any-return]
				lambda kernel, x1, x2: kernel(x1, x2),
				in_axes=(
					self.batch_in_axes,
					self.batch_over_inputs,
					self.batch_over_inputs if x2 is not None else None,
				),
			)(self.inner, x1, x2)
		else:
			if self.batch_size == 1:
				# If batch size is 1, we can just call the inner kernel without repeating
				return self.inner(x1, x2)[None, ...]  # Add batch dimension

			# Repeat the same matrix when all hyperparameters and inputs are shared
			return jnp.repeat(
				jnp.expand_dims(self.inner(x1, x2), 0),
				self.batch_size,
				axis=0
			)

	def __str__(self):
		# just str of the inner kernel, as the batch info is in the parameters of the inner kernel
		return f"{self.inner}"



"""
Configuration system for kernax.

This module provides a global configuration object that can be accessed via `kernax.config`.
Configuration options can be set globally or temporarily within a context manager.

IMPORTANT: parameter_transform must be set BEFORE creating any kernels, as JIT compilation
caches prevent dynamic changes. It cannot be used with the set_config() context manager.

Example:
	>>> import kernax
	>>> # Set parameter transform before creating kernels
	>>> kernax.config.parameter_transform = "softplus"
	>>> kernel = kernax.SEKernel(length_scale=1.0)
"""

from __future__ import annotations

import warnings
from contextvars import ContextVar
from typing import Any, Literal

# Valid parameter transform modes
ParameterTransform = Literal["identity", "exp", "softplus"]

# Context variable to store thread-local configuration
_config_context: ContextVar[dict[str, Any] | None] = ContextVar("kernax_config", default=None)


class Config:
	"""
	Global configuration object for kernax.

	This class manages global and thread-local configuration options for the library.
	Configuration can be modified globally or temporarily within a context manager.

	IMPORTANT: parameter_transform must be set BEFORE creating any kernels, as changing it
	after kernel instantiation would cause inconsistencies with JIT-compiled functions.

	Attributes:
		parameter_transform: Transform to apply to positive-constrained parameters.
			- "identity": No transform (parameters stay in constrained space)
			- "exp": Exponential transform (log-exp trick)
			- "softplus": Softplus transform (log(1 + exp(x)))
			Default: "identity"
			NOTE: Must be set before creating any kernels!
	"""

	def __init__(self):
		# Default configuration values
		self._global_config: dict[str, Any] = {
			"parameter_transform": "identity",
		}
		# Track whether any kernels have been instantiated
		self._kernels_instantiated = False
		# Allow unsafe changes (for testing purposes only)
		self._allow_unsafe_changes = False

	@property
	def parameter_transform(self) -> ParameterTransform:
		"""
		Get the current parameter transform mode.

		Returns:
			The current parameter transform: "identity", "exp", or "softplus"
		"""
		return self._get_value("parameter_transform")  # type: ignore[no-any-return]

	@parameter_transform.setter
	def parameter_transform(self, value: ParameterTransform) -> None:
		"""
		Set the parameter transform mode globally.

		IMPORTANT: This must be set BEFORE creating any kernels. Once a kernel has been
		instantiated, changing this value would cause inconsistencies with JIT-compiled
		functions.

		Args:
			value: Parameter transform mode ("identity", "exp", or "softplus")

		Raises:
			ValueError: If value is not a valid transform mode
			RuntimeError: If trying to change after kernels have been instantiated

		Example:
			>>> import kernax
			>>> kernax.config.parameter_transform = "softplus"  # Before kernels
			>>> kernel = kernax.SEKernel(length_scale=1.0)
			>>> # kernax.config.parameter_transform = "exp"  # Would raise RuntimeError!
		"""
		# Check if kernels have been instantiated
		if self._kernels_instantiated and not self._allow_unsafe_changes:
			raise RuntimeError(
				"Cannot change parameter_transform after kernels have been instantiated.\n"
				"Reason: JIT-compiled functions cache the transform value, causing "
				"inconsistent behavior.\n"
				"Solution: Set parameter_transform before creating kernels, or call\n"
				"  kernax.config.unsafe_reset() to reset (invalidates existing kernels)."
			)

		# Validate value
		valid_transforms: tuple[ParameterTransform, ...] = ("identity", "exp", "softplus")
		if value not in valid_transforms:
			raise ValueError(
				f"Invalid parameter_transform: {value}. "
				f"Must be one of {valid_transforms}"
			)

		self._global_config["parameter_transform"] = value

	def _get_value(self, key: str) -> Any:
		"""
		Get a configuration value, checking thread-local context first, then global.

		Args:
			key: Configuration key to retrieve

		Returns:
			The configuration value
		"""
		context_config = _config_context.get()
		if context_config is not None and key in context_config:
			return context_config[key]
		return self._global_config[key]

	def get_all(self) -> dict[str, Any]:
		"""
		Get all current configuration values (global + thread-local overrides).

		Returns:
			Dictionary of all configuration values
		"""
		config = self._global_config.copy()
		context_config = _config_context.get()
		if context_config is not None:
			config.update(context_config)
		return config

	def set_config(self, **kwargs: Any) -> ConfigContext:
		"""
		Create a context manager for temporary configuration changes.

		This allows you to temporarily override configuration values within a
		specific scope. Changes are automatically reverted when exiting the context.

		NOTE: parameter_transform CANNOT be used with this context manager due to
		JIT compilation caching. It must be set globally before creating kernels.

		Args:
			**kwargs: Configuration values to set temporarily

		Returns:
			A context manager that applies the configuration changes

		Example:
			>>> # This would work if we had other config options:
			>>> # with kernax.config.set_config(some_other_option=value):
			>>> #     ...

		Raises:
			ValueError: If any configuration key or value is invalid, or if
				parameter_transform is specified
		"""
		# Block parameter_transform in context manager
		if "parameter_transform" in kwargs:
			raise ValueError(
				"parameter_transform cannot be used with set_config() context manager.\n"
				"Reason: JIT compilation caches prevent dynamic changes.\n"
				"Solution: Set it globally before creating kernels:\n"
				"  kernax.config.parameter_transform = 'exp'"
			)

		# Validate all keys and values before creating context
		for key, _value in kwargs.items():
			if key not in self._global_config:
				raise ValueError(
					f"Unknown configuration key: {key}. "
					f"Valid keys: {list(self._global_config.keys())}"
				)

		return ConfigContext(kwargs)

	def _mark_module_instantiated(self) -> None:
		"""
		Mark that a kernel has been instantiated.

		This is called automatically by AbstractKernel.__init__() and locks the
		parameter_transform setting to prevent inconsistencies with JIT-compiled code.

		Internal use only - should not be called by users.
		"""
		self._kernels_instantiated = True

	def reset(self) -> None:
		"""
		Reset all global configuration values to their defaults.

		This does not affect thread-local context overrides currently in effect.

		NOTE: This does NOT reset the kernel instantiation lock. If kernels have been
		created, you cannot change parameter_transform even after reset(). Use
		unsafe_reset() if you need to reset everything (use with caution).
		"""
		self._global_config = {
			"parameter_transform": "identity",
		}

	def unsafe_reset(self) -> None:
		"""
		Reset configuration AND clear the kernel instantiation lock.

		WARNING: This is unsafe! After calling this, any existing kernels will behave
		inconsistently because their JIT-compiled code may use different transform
		settings than newly created kernels.

		Use this only when:
		- You are certain no existing kernels will be used
		- You are in an interactive session and want to start fresh
		- You are writing tests

		After calling this, you should delete all existing kernel instances before
		creating new ones.
		"""
		warnings.warn(
			"unsafe_reset() clears the kernel instantiation lock. "
			"Any existing kernels will behave inconsistently if used after this call. "
			"Delete all kernel instances before creating new ones.",
			RuntimeWarning,
			stacklevel=2,
		)
		self._kernels_instantiated = False
		self._allow_unsafe_changes = False
		self.reset()

	def __repr__(self) -> str:
		"""String representation showing current configuration."""
		config = self.get_all()
		config_str = ", ".join(f"{k}={v!r}" for k, v in sorted(config.items()))
		return f"Config({config_str})"


class ConfigContext:
	"""
	Context manager for temporary configuration changes.

	This class should not be instantiated directly. Use `config.set_config()` instead.
	"""

	def __init__(self, overrides: dict[str, Any]):
		"""
		Initialize the context manager.

		Args:
			overrides: Configuration values to override temporarily
		"""
		self.overrides = overrides
		self.token: Any = None

	def __enter__(self) -> ConfigContext:
		"""Enter the context and apply configuration overrides."""
		# Get current context config and merge with new overrides
		current_context = _config_context.get()
		if current_context is None:
			current_context = {}
		else:
			current_context = current_context.copy()
		current_context.update(self.overrides)
		self.token = _config_context.set(current_context)
		return self

	def __exit__(self, exc_type, exc_val, exc_tb) -> None:
		"""Exit the context and restore previous configuration."""
		if self.token is not None:
			_config_context.reset(self.token)


# Global configuration instance
config = Config()



""" Defines distances/comparisons between input vectors, used in different families of kernels. """

import jax.numpy as jnp
from jax import Array, jit


@jit
def euclidean_distance(x1: Array, x2: Array) -> Array:
	"""
	Compute the Euclidean distance between two vectors.

	:param x1: First input vector.
	:param x2: Second input vector.
	:return: Euclidean distance between x1 and x2.
	"""
	return jnp.linalg.norm(x1 - x2)  # type: ignore[no-any-return]

@jit
def squared_euclidean_distance(x1: Array, x2: Array) -> Array:
	"""
	Compute the squared Euclidean distance between two vectors.

	:param x1: First input vector.
	:param x2: Second input vector.
	:return: Squared Euclidean distance between x1 and x2.
	"""
	return jnp.sum((x1 - x2)**2)

@jit
def manhattan_distance(x1: Array, x2: Array) -> Array:
	"""
	Compute the Manhattan (L1) distance between two vectors.

	:param x1: First input vector.
	:param x2: Second input vector.
	:return: Manhattan distance between x1 and x2.
	"""
	return jnp.sum(jnp.abs(x1 - x2))

@jit
def chebyshev_distance(x1: Array, x2: Array) -> Array:
	"""
	Compute the Chebyshev (L∞) distance between two vectors.

	:param x1: First input vector.
	:param x2: Second input vector.
	:return: Chebyshev distance between x1 and x2.
	"""
	return jnp.max(jnp.abs(x1 - x2))

@jit
def cosine_distance(x1: Array, x2: Array) -> Array:
	"""
	Compute the Cosine distance between two vectors.

	:param x1: First input vector.
	:param x2: Second input vector.
	:return: Cosine distance between x1 and x2.
	"""
	dot_product = jnp.dot(x1, x2)
	norm_x1 = jnp.linalg.norm(x1)
	norm_x2 = jnp.linalg.norm(x2)
	return 1.0 - (dot_product / (norm_x1 * norm_x2))  # type: ignore[no-any-return]

@jit
def minkowski_distance(x1: Array, x2: Array, p: float) -> Array:
	"""
	Compute the Minkowski distance between two vectors.

	:param x1: First input vector.
	:param x2: Second input vector.
	:param p: Order of the norm (p >= 1).
	:return: Minkowski distance between x1 and x2.
	"""
	return jnp.sum(jnp.abs(x1 - x2) ** p) ** (1 / p)

@jit
def hamming_distance(x1: Array, x2: Array) -> Array:
	"""
	Compute the Hamming distance between two vectors.

	:param x1: First input vector (binary or categorical).
	:param x2: Second input vector (binary or categorical).
	:return: Hamming distance between x1 and x2.
	"""
	return jnp.sum(x1 != x2)

@jit
def jaccard_distance(x1: Array, x2: Array) -> Array:
	"""
	Compute the Jaccard distance between two binary vectors.

	:param x1: First input vector (binary).
	:param x2: Second input vector (binary).
	:return: Jaccard distance between x1 and x2.
	"""
	intersection = jnp.sum(jnp.minimum(x1, x2))
	union = jnp.sum(jnp.maximum(x1, x2))
	return 1.0 - (intersection / union)

@jit
def mahalanobis_distance(x1: Array, x2: Array, inv_cov_matrix: Array) -> Array:
	"""
	Compute the Mahalanobis distance between two vectors.

	:param x1: First input vector.
	:param x2: Second input vector.
	:param inv_cov_matrix: Inverse of the covariance matrix.
	:return: Mahalanobis distance between x1 and x2.
	"""
	diff = x1 - x2
	return jnp.sqrt(jnp.dot(jnp.dot(diff.T, inv_cov_matrix), diff))

@jit
def canberra_distance(x1: Array, x2: Array) -> Array:
	"""
	Compute the Canberra distance between two vectors.

	:param x1: First input vector.
	:param x2: Second input vector.
	:return: Canberra distance between x1 and x2.
	"""
	return jnp.sum(jnp.abs(x1 - x2) / (jnp.abs(x1) + jnp.abs(x2) + 1e-10))  # Added small constant to avoid division by zero

@jit
def bray_curtis_distance(x1: Array, x2: Array) -> Array:
	"""
	Compute the Bray-Curtis distance between two vectors.

	:param x1: First input vector.
	:param x2: Second input vector.
	:return: Bray-Curtis distance between x1 and x2.
	"""
	return jnp.sum(jnp.abs(x1 - x2)) / jnp.sum(jnp.abs(x1 + x2) + 1e-10)  # Added small constant to avoid division by zero

@jit
def correlation_distance(x1: Array, x2: Array) -> Array:
	"""
	Compute the Correlation distance between two vectors.

	:param x1: First input vector.
	:param x2: Second input vector.
	:return: Correlation distance between x1 and x2.
	"""
	x1_mean = jnp.mean(x1)
	x2_mean = jnp.mean(x2)
	numerator = jnp.sum((x1 - x1_mean) * (x2 - x2_mean))
	denominator = jnp.sqrt(jnp.sum((x1 - x1_mean) ** 2) * jnp.sum((x2 - x2_mean) ** 2))
	return 1.0 - (numerator / (denominator + 1e-10))  # Added small constant to avoid division by zero

@jit
def dot_product(x1: Array, x2: Array) -> Array:
	"""
	Compute the Dot Product distance between two vectors.

	:param x1: First input vector.
	:param x2: Second input vector.
	:return: Dot Product between x1 and x2.
	"""
	return x1.T @ x2



from __future__ import annotations

from typing import Optional, Tuple

import jax.numpy as jnp
from equinox import error_if, field, filter_jit
from jax import Array

from .WrapperModule import WrapperModule


class ActiveDimsModule(WrapperModule):
	"""
	Wrapper module to select active dimensions from the inputs before passing them to the inner module.

	NOTE: This module *must* be the outer-most wrapper (aka it shouldn't be wrapped inside another one).
	If you use a kernel that has HPs specific to *input dimensions* (like an ARDKernel), make sure you
	instantiate it with HPs only for the active dimensions.
	"""

	active_dims: Tuple[int, ...] = field(static=True)

	def __init__(self, inner, active_dims: Tuple[int, ...], **kwargs):
		super().__init__(inner=inner, **kwargs)
		self.active_dims = tuple(int(dim) for dim in active_dims)

	@filter_jit
	def __call__(self, x1: Array, x2: Optional[Array] = None) -> Array:
		x1 = error_if(x1, jnp.any(jnp.array(self.active_dims) >= x1.shape[-1]),
		              "active_dims contains indices out of bounds for x1.")

		if x2 is None:
			return self.inner(x1[..., self.active_dims])
		return self.inner(x1[..., self.active_dims], x2[..., self.active_dims])


import equinox as eqx
from equinox import filter_jit
from jax import Array
from jax import numpy as jnp

from ..AbstractKernel import AbstractKernel
from ..distances import euclidean_distance
from .StationaryKernel import StaticStationaryKernel


# Matern 5/2 Kernel defined in Rasmussen and Williams (2006), section 4.2
class StaticMatern52Kernel(StaticStationaryKernel):
	distance_func = euclidean_distance

	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute the Matern 5/2 kernel covariance value between two vectors.

		:param kern: the kernel to use, containing hyperparameters (length_scale)
		:param x1: scalar array
		:param x2: scalar array
		:return: scalar array
		"""
		kern = eqx.combine(kern)
		r = cls.distance_func(x1, x2)
		sqrt5_r_div_l = (jnp.sqrt(5) * r) / kern.length_scale  # type: ignore[attr-defined]
		return (1.0 + sqrt5_r_div_l + (5.0 / 3.0) * (r / kern.length_scale) ** 2) * jnp.exp(  # type: ignore[no-any-return,attr-defined]
			-sqrt5_r_div_l
		)


class Matern52Kernel(AbstractKernel):
	"""
	Matern 5/2 Kernel

	The length_scale parameter is always positive. Internally, it may be stored in an
	unconstrained space depending on the global configuration.
	"""

	_raw_length_scale: Array = eqx.field(converter=jnp.asarray)
	static_class = StaticMatern52Kernel

	def __init__(self, length_scale, **kwargs):
		"""
		Initialize the Matern 5/2 kernel with a length scale parameter.

		Args:
			length_scale: Length scale parameter (must be positive)

		Raises:
			ValueError: If length_scale is not positive
		"""
		# Validate positivity
		length_scale = jnp.array(length_scale)
		length_scale = eqx.error_if(
			length_scale, jnp.any(length_scale <= 0), "length_scale must be positive."
		)

		# Initialize parent (locks config)
		super().__init__(**kwargs)

		# Transform to unconstrained space
		from ..transforms import to_unconstrained

		self._raw_length_scale = to_unconstrained(jnp.asarray(length_scale))

	@property
	def length_scale(self) -> Array:
		"""Get the length scale in constrained space (always positive)."""
		from ..transforms import to_constrained

		return to_constrained(self._raw_length_scale)



import equinox as eqx
from jax import Array
from jax import numpy as jnp

from .ConstantKernel import AbstractKernel, StaticConstantKernel
from ..transforms import to_constrained, to_unconstrained


class VarianceKernel(AbstractKernel):
	"""
	Variance kernel that returns a constant value everywhere. Used to multiply with other kernels.

	This kernel is functionally equivalent to a ConstantKernel but the implementation differs in
	two ways:
	1) The "value" parameter is renamed "variance" parameter, allowing easier modification when
	combining with another ConstantKernel (e.g. WhiteNoiseKernel)
	2) It forces a positivity constraint on the variance parameter

	"""

	_raw_variance: Array = eqx.field(converter=jnp.asarray)

	@property
	def variance(self) -> Array:
		return to_constrained(self._raw_variance)

	@property
	def value(self):  # As we use the StaticConstantKernel pairwise_cov, we need to provide a 'value' property that returns the variance
		return to_constrained(self._raw_variance)

	static_class = StaticConstantKernel

	def __init__(self, variance=1.0, **kwargs):
		# Assert noise is positive
		variance = jnp.array(variance)
		variance = eqx.error_if(variance, jnp.any(variance <= 0), "variance must be positive.")

		super().__init__(**kwargs)

		# Transform to unconstrained space
		self._raw_variance = to_unconstrained(jnp.asarray(variance))




from __future__ import annotations

from typing import Optional

from equinox import filter_jit
from jax import Array

from .OperatorModule import OperatorModule


class ProductModule(OperatorModule):
	"""Operator module that multiplies the outputs of two sub-modules."""

	@filter_jit
	def __call__(self, x1: Array, x2: Optional[Array] = None) -> Array:
		if x2 is None:
			return self.left(x1) * self.right(x1)
		return self.left(x1, x2) * self.right(x1, x2)

	def __str__(self):
		if self.right.__class__.__name__ == "NegModule":
			return f"{self.left} * ({self.right})"
		return f"{self.left} * {self.right}"


import equinox as eqx
from equinox import filter_jit
from jax import Array
from jax import numpy as jnp

from ..AbstractKernel import AbstractKernel
from ..distances import euclidean_distance
from .StationaryKernel import StaticStationaryKernel


class StaticPeriodicKernel(StaticStationaryKernel):
	distance_func = euclidean_distance

	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute the periodic kernel covariance value between two vectors.

		:param kern: the kernel to use, containing hyperparameters (length_scale, variance, period).
		:param x1: scalar array
		:param x2: scalar array
		:return: covariance value (scalar)
		"""
		kern = eqx.combine(kern)
		dist = cls.distance_func(x1, x2)

		return kern.variance * jnp.exp(  # type: ignore[no-any-return,attr-defined]
			-2 * jnp.sin(jnp.pi * dist / kern.period) ** 2 / kern.length_scale**2  # type: ignore[attr-defined]
		)


class PeriodicKernel(AbstractKernel):
	"""
	Periodic Kernel

	All parameters (length_scale, variance, period) are always positive.
	Internally, they may be stored in unconstrained space.
	"""

	_raw_length_scale: Array = eqx.field(converter=jnp.asarray)
	_raw_variance: Array = eqx.field(converter=jnp.asarray)
	_raw_period: Array = eqx.field(converter=jnp.asarray)
	static_class = StaticPeriodicKernel

	def __init__(self, length_scale, variance, period, **kwargs):
		"""
		Initialize the Periodic kernel.

		Args:
			length_scale: length scale parameter (ℓ, must be positive)
			variance: variance parameter (σ², must be positive)
			period: period parameter (p, must be positive)

		Raises:
			ValueError: If any parameter is not positive
		"""
		# Validate positivity
		length_scale = jnp.array(length_scale)
		variance = jnp.array(variance)
		period = jnp.array(period)

		length_scale = eqx.error_if(
			length_scale, jnp.any(length_scale <= 0), "length_scale must be positive."
		)
		variance = eqx.error_if(variance, jnp.any(variance <= 0), "variance must be positive.")
		period = eqx.error_if(period, jnp.any(period <= 0), "period must be positive.")

		# Initialize parent (locks config)
		super().__init__(**kwargs)

		# Transform to unconstrained space
		from ..transforms import to_unconstrained

		self._raw_length_scale = to_unconstrained(jnp.asarray(length_scale))
		self._raw_variance = to_unconstrained(jnp.asarray(variance))
		self._raw_period = to_unconstrained(jnp.asarray(period))

	@property
	def length_scale(self) -> Array:
		"""Get the length scale in constrained space (always positive)."""
		from ..transforms import to_constrained

		return to_constrained(self._raw_length_scale)

	@property
	def variance(self) -> Array:
		"""Get the variance in constrained space (always positive)."""
		from ..transforms import to_constrained

		return to_constrained(self._raw_variance)

	@property
	def period(self) -> Array:
		"""Get the period in constrained space (always positive)."""
		from ..transforms import to_constrained

		return to_constrained(self._raw_period)




from ..AbstractKernel import StaticAbstractKernel
from ..distances import dot_product


class StaticDotProductKernel(StaticAbstractKernel):
	"""
	Super-class for every kernel that uses the dot product between input vectors.

	This allows to change the distance function used in child classes.
	The default metric is the dot product.
	"""

	distance_func = dot_product



import equinox as eqx
from equinox import filter_jit
from jax import Array
from jax import numpy as jnp

from ..AbstractKernel import AbstractKernel
from .DotProductKernel import StaticDotProductKernel


class StaticSigmoidKernel(StaticDotProductKernel):
	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute the sigmoid kernel covariance value between two vectors.

		Formula: tanh(α⟨x, x'⟩ + c)

		:param kern: kernel instance containing the hyperparameters (alpha, constant)
		:param x1: scalar array
		:param x2: scalar array
		:return: scalar array
		"""
		kern = eqx.combine(kern)
		dp = cls.distance_func(x1, x2)
		return jnp.tanh(kern.alpha * dp + kern.constant)  # type: ignore[attr-defined]


class SigmoidKernel(AbstractKernel):
	"""
	Sigmoid (Hyperbolic Tangent) Kernel

	Formula: tanh(α⟨x, x'⟩ + c)

	Parameter alpha must be positive.
	Parameter constant can be any real value.
	"""

	_raw_alpha: Array = eqx.field(converter=jnp.asarray)
	constant: Array = eqx.field(converter=jnp.asarray)
	static_class = StaticSigmoidKernel

	def __init__(self, alpha: float = 1.0, constant: float = 0.0):
		"""
		Initialize the Sigmoid kernel.

		Args:
			alpha: Scale factor (must be positive). Controls the steepness.
			constant: Independent term (can be any real value). Shifts the activation.

		Raises:
			ValueError: If alpha is not positive
		"""
		# Validate alpha positivity
		alpha_array = jnp.array(alpha)
		alpha_array = eqx.error_if(alpha_array, jnp.any(alpha_array <= 0), "alpha must be positive.")

		# Initialize parent (locks config)
		super().__init__()

		# Transform alpha to unconstrained space
		from ..transforms import to_unconstrained

		self._raw_alpha = to_unconstrained(jnp.asarray(alpha_array))

		# constant can be any value, no transformation needed
		self.constant = jnp.asarray(constant)

	@property
	def alpha(self) -> Array:
		"""Get the alpha parameter in constrained space (always positive)."""
		from ..transforms import to_constrained

		return to_constrained(self._raw_alpha)



"""
Parameter transformation utilities for kernax.

This module provides functions to transform parameters between constrained and
unconstrained spaces, enabling stable optimization of positive-constrained parameters.

The transformation used depends on the global config.parameter_transform setting:
- "identity": No transformation (parameters stay in constrained space)
- "exp": Exponential transform (log-exp trick)
- "softplus": Softplus transform (numerically stable near zero)
"""

from __future__ import annotations

import jax.numpy as jnp
from jax import Array


def to_unconstrained(value: Array) -> Array:
	"""
	Transform a positive parameter to unconstrained space.

	This function converts a parameter from constrained (positive) space to
	unconstrained (real) space based on the current config.parameter_transform setting.

	Args:
		value: Parameter value in constrained space (must be positive)

	Returns:
		Parameter value in unconstrained space

	Example:
		>>> import kernax
		>>> import jax.numpy as jnp
		>>> kernax.config.parameter_transform = "exp"
		>>> unconstrained = kernax.transforms.to_unconstrained(jnp.array(2.0))
		>>> # unconstrained ≈ 0.693 (log(2.0))
	"""
	from .config import config

	if config.parameter_transform == "identity":
		# No transformation
		return value
	elif config.parameter_transform == "exp":
		# Inverse of exp is log
		return jnp.log(value)
	elif config.parameter_transform == "softplus":
		# Inverse of softplus: log(exp(x) - 1)
		return jnp.log(jnp.exp(value) - 1.0)
	else:
		raise ValueError(f"Unknown parameter_transform: {config.parameter_transform}")


def to_constrained(value: Array) -> Array:
	"""
	Transform a parameter from unconstrained space to positive constrained space.

	This function converts a parameter from unconstrained (real) space back to
	constrained (positive) space based on the current config.parameter_transform setting.

	Args:
		value: Parameter value in unconstrained space

	Returns:
		Parameter value in constrained space (always positive)

	Example:
		>>> import kernax
		>>> import jax.numpy as jnp
		>>> kernax.config.parameter_transform = "exp"
		>>> constrained = kernax.transforms.to_constrained(jnp.array(0.693))
		>>> # constrained ≈ 2.0 (exp(0.693))
	"""
	from .config import config

	if config.parameter_transform == "identity":
		# No transformation
		return value
	elif config.parameter_transform == "exp":
		# Exponential ensures positivity
		return jnp.exp(value)
	elif config.parameter_transform == "softplus":
		# Softplus: log(1 + exp(x))
		return jnp.log(1.0 + jnp.exp(value))
	else:
		raise ValueError(f"Unknown parameter_transform: {config.parameter_transform}")



import equinox as eqx
import jax.numpy as jnp
import jax.tree_util as jtu
from equinox import filter_jit
from jax import Array, vmap

from ..AbstractKernel import AbstractKernel
from .WrapperModule import WrapperModule


class BlockKernel(WrapperModule):
	"""
	Wrapper kernel to build block covariance matrices using any kernel.

	A basic kernel usually works on inputs of shape (N, I), and produces covariance matrices of shape (N, N).

	Wrapped inside a block kernel, they can either:
	- still work on inputs of shape (N, I), but produce covariance matrices of shape (B*N, B*N), where B is the number of blocks. This is useful when the hyperparameters are distinct to blocks, i.e. each sub-matrix has its own set of hyperparameters.
	- or work on inputs of shape (B, N, I), producing covariance matrices of shape (B*N, B*N). This is useful when inputs are different for each block, regardless of whether the hyperparameters are shared between blocks or not.

	This class uses vmap to vectorize the kernel computation of each block, then resize the result into a block matrix.

	N.b: when hyper-parameters are different from one block to another, the inner must do its computations on a kernel instance containing two pairs of hyper-parameters (one per block dimension). For example, a blocked SEKernel would receive two distinct lengthscales and variances, one specific to the row of the block, the other specific to the column of the block. Not all kernels are designed to handle such cases, so be careful when using this class with varying hyper-parameters. A good example of a kernel designed for this is the `MultiFeatureKernel`.
	"""

	inner: AbstractKernel = eqx.field()
	nb_blocks: int = eqx.field(static=True)
	block_in_axes: bool = eqx.field(static=True)
	block_over_inputs: int | None = eqx.field(static=True)

	def __init__(self, inner, nb_blocks, block_in_axes=None, block_over_inputs=True, **kwargs):
		"""
		:param inner: the kernel to wrap, must be an instance of AbstractKernel associated to a static_class that can handle two pairs of hyper-parameters when needed.
		:param nb_blocks: the number of blocks, B
		:param block_in_axes: a pytree indicating which hyperparameters change across blocks.
								If 0, the hyperparameter changes across blocks of the matrix.
								If None, the hyperparameter is shared across all blocks.
		:param block_over_inputs: whether to expect inputs of shape (B, N, I) (True) or (N, I) (False)

		N.b: the result of this kernel is not always a valid covariance matrix!
		Usually, you want to use this kernel with an appropriate inner, calculating a function where two hyper-parameters have symmetric roles.
		A good example is a multi-output (convolutional) kernel in GPs, which usually have two distinct length_scales (and variances) depending on which output dimension is considered.
		"""
		# Initialize the WrapperKernel
		super().__init__(inner=inner, **kwargs)
		self.nb_blocks = nb_blocks

		# Default: all array hyperparameters are shared (None for all array leaves)
		if block_in_axes is None:
			# Extract only array leaves and map them to None
			self.block_in_axes = jtu.tree_map(lambda _: None, inner)
		elif block_in_axes == 0:
			# All hyperparameters are blocked
			self.block_in_axes = jtu.tree_map(lambda _: 0, inner)
		else:
			self.block_in_axes = block_in_axes

		self.block_over_inputs = 0 if block_over_inputs else None

		# Add block dimension to parameters where batch_in_axes is 0
		self.inner = jtu.tree_map(
			lambda param, block_in_ax: (
				param if block_in_ax is None else jnp.repeat(param[None, ...], nb_blocks, axis=0)
			),
			self.inner,
			self.block_in_axes,
		)

	@filter_jit
	def __call__(self, x1: Array, x2: None | Array = None) -> Array:
		"""
		Compute the kernel over batched inputs using vmap.

		Args:
				x1: Input of shape (B, ..., N, I)
				x2: Optional second input of shape (B, ..., M, I)

		Returns:
				Kernel block-matrix of appropriate shape
		"""
		x2 = x1 if x2 is None else x2

		# Check if we can use vmap (at least one axis is not None)
		can_use_vmap = (
			not jtu.tree_all(jtu.tree_map(lambda k: k is None, self.block_in_axes))
			or self.block_over_inputs is not None
		)

		if can_use_vmap:
			# Use vmap when we have blocked hyperparameters or blocked inputs
			rows, cols = jnp.triu_indices(self.nb_blocks)

			pair_kernel = jtu.tree_map(
				lambda param, block_in_ax:
				jnp.swapaxes(jnp.array([param[rows], param[cols]]), 0, 1)
				if block_in_ax == 0
				else param,
				self.inner,
				self.block_in_axes,
			)

			x1_indexed = x1[rows] if self.block_over_inputs == 0 else x1
			x2_indexed = x2[cols] if self.block_over_inputs == 0 else x2

			# vmap over the block dimension of inner and inputs
			# Each block element gets its own version of inner with corresponding hyperparameters
			return self.symmetric_blocks_to_matrix(  # type: ignore[no-any-return]
				vmap(
					lambda kernel, x_1, x_2: kernel(x_1, x_2),
					in_axes=(
						self.block_in_axes,
						self.block_over_inputs,
						self.block_over_inputs,
					),
				)(pair_kernel, x1_indexed, x2_indexed)
			)
		else:
			# All hyperparameters and inputs are shared: create a block matrix with identical blocks
			single_block = self.inner(x1, x2)
			# Tile the block to create the full block matrix
			return jnp.tile(single_block, (self.nb_blocks, self.nb_blocks))

	@filter_jit
	def symmetric_blocks_to_matrix(self, flat_blocks):
		"""
		Rebuilds a symmetric matrix from its unique blocks (upper triangle).

		Args:
			flat_blocks: Tensor (T, H, W) where T is a triangular number.
						 Expected order: (0,0), (0,1), (0,2)... (row-major upper)
		"""
		# 2. Créer la grille de blocs vide (B, B, H, W)
		grid = jnp.zeros(
			(self.nb_blocks, self.nb_blocks, flat_blocks.shape[-2], flat_blocks.shape[-1]),
			dtype=flat_blocks.dtype,
		)

		# 3. Récupérer les indices du triangle supérieur
		# ex: rows=[0,0,1], cols=[0,1,1] pour B=2
		rows, cols = jnp.triu_indices(self.nb_blocks)

		# 4. Remplir le triangle supérieur
		grid = grid.at[rows, cols].set(flat_blocks)

		# 5. Remplir le triangle inférieur par symétrie
		# On prend le bloc en (rows, cols), on le transpose (swapaxes -1, -2)
		# et on le place en (cols, rows).
		# Note : Sur la diagonale (rows==cols), cela transpose le bloc sur lui-même,
		# ce qui est correct car un bloc diagonal d'une matrice symétrique est lui-même symétrique.
		grid = grid.at[cols, rows].set(grid[rows, cols].swapaxes(-1, -2))

		# 6. Assemblage final (Même logique que précédemment)
		return grid.swapaxes(1, 2).reshape(  # (RowBlock, H, ColBlock, W)
			self.nb_blocks * flat_blocks.shape[-2], self.nb_blocks * flat_blocks.shape[-1]
		)

	def __str__(self):
		return f"Block{self.inner}"



from __future__ import annotations

import equinox as eqx

from ..module import AbstractModule


def _to_constant(value, reference_module):
	"""Convert a non-Module value to the appropriate Constant type based on the other operand."""
	from ..AbstractKernel import AbstractKernel
	from ..AbstractMean import AbstractMean

	if isinstance(reference_module, AbstractMean):
		from ..means.ConstantMean import ConstantMean
		return ConstantMean(constant=value)
	else:
		from ..other.ConstantKernel import ConstantKernel
		return ConstantKernel(value=value)


class OperatorModule(AbstractModule):
	"""Base class for modules that apply an operation on the outputs of two sub-modules."""

	left: AbstractModule = eqx.field()
	right: AbstractModule = eqx.field()

	def __init__(self, left, right, **kwargs):
		if not isinstance(left, AbstractModule):
			left = _to_constant(left, right)
		if not isinstance(right, AbstractModule):
			right = _to_constant(right, left)

		super().__init__(**kwargs)
		self.left = left
		self.right = right

	def replace(self, **kwargs):
		operator_kwargs = {}
		param_kwargs = {}

		for k, v in kwargs.items():
			if k in ["left", "right"]:
				operator_kwargs[k] = v
			else:
				param_kwargs[k] = v

		result = self

		if operator_kwargs:
			result = super().replace(**operator_kwargs)

		if param_kwargs:
			try:
				new_left = result.left.replace(**param_kwargs)
			except (AttributeError, TypeError):
				new_left = result.left

			try:
				new_right = result.right.replace(**param_kwargs)
			except (AttributeError, TypeError):
				new_right = result.right

			result = eqx.tree_at(lambda s: (s.left, s.right), result, (new_left, new_right))

		return result



from __future__ import annotations

from typing import Optional

from equinox import filter_jit
from jax import Array

from .OperatorModule import OperatorModule


class SumModule(OperatorModule):
	"""Operator module that sums the outputs of two sub-modules."""

	@filter_jit
	def __call__(self, x1: Array, x2: Optional[Array] = None) -> Array:
		if x2 is None:
			return self.left(x1) + self.right(x1)
		return self.left(x1, x2) + self.right(x1, x2)

	def __str__(self):
		if self.right.__class__.__name__ == "NegModule":
			return f"{self.left} - {self.right.inner}"  # type: ignore[attr-defined]
		return f"{self.left} + {self.right}"



from ..AbstractKernel import StaticAbstractKernel
from ..distances import squared_euclidean_distance


class StaticStationaryKernel(StaticAbstractKernel):
	"""
	Super-class for every stationary/isotropic kernel.

	The isotropic property depends only on the distance function used. You can check available
	distance function in `kernax/distances.py`.

	The default distance is the squared Euclidean distance, but it can be overridden in child classes.
	"""

	distance_func = squared_euclidean_distance



class StaticZeroMean(StaticAbstractMean):
	@classmethod
	@filter_jit
	def scalar_mean(cls, mean: AbstractMean, x: Array) -> Array:
		return jnp.array(0.)

class ZeroMean(AbstractMean):
	static_class = StaticZeroMean


class StaticConstantMean(StaticAbstractMean):
	@classmethod
	@filter_jit
	def scalar_mean(cls, mean: AbstractMean, x: Array) -> Array:
		return mean.constant

class ConstantMean(AbstractMean):
	constant: Array = eqx.field(converter=jnp.asarray)
	static_class = StaticConstantMean

	def __init__(self, constant=0., **kwargs):
		super().__init__(**kwargs)
		self.constant = jnp.asarray(constant)



class StaticLinearMean(StaticAbstractMean):
	@classmethod
	@filter_jit
	def scalar_mean(cls, mean: AbstractMean, x: Array) -> Array:
		return mean.slope * x

class LinearMean(AbstractMean):
	slope: Array = eqx.field(converter=jnp.asarray)
	static_class = StaticLinearMean

	def __init__(self, slope=0., **kwargs):
		super().__init__(**kwargs)
		self.slope = jnp.asarray(slope)

class StaticAffineMean(StaticAbstractMean):
	@classmethod
	@filter_jit
	def scalar_mean(cls, mean: AbstractMean, x: Array) -> Array:
		return mean.slope * x + mean.intercept

class AffineMean(AbstractMean):
	slope: Array = eqx.field(converter=jnp.asarray)
	intercept: Array = eqx.field(converter=jnp.asarray)
	static_class = StaticAffineMean

	def __init__(self, slope=0., intercept=0., **kwargs):
		super().__init__(**kwargs)
		self.slope = jnp.asarray(slope)
		self.intercept = jnp.asarray(intercept)



import equinox as eqx
from jax import Array
from jax import numpy as jnp

from ..engines import SafeDiagonalEngine
from .ConstantKernel import AbstractKernel, StaticConstantKernel
from ..transforms import to_constrained, to_unconstrained


class WhiteNoiseKernel(AbstractKernel):
	"""
	White noise kernel that returns a constant value only on the diagonal.

	This kernel is equivalent to a ConstantKernel with a SafeDiagonalEngine computation engine.
	It returns the noise value when inputs are equal, and 0 otherwise, resulting in a diagonal
	covariance matrix.
	"""

	_raw_noise: Array = eqx.field(converter=jnp.asarray)

	@property
	def noise(self) -> Array:
		return to_constrained(self._raw_noise)

	@property
	def value(self):  # As we use the StaticConstantKernel pairwise_cov, we need to provide a 'value' property that returns the noise level
		return to_constrained(self._raw_noise)

	static_class = StaticConstantKernel

	def __init__(self, noise=1.0, **kwargs):
		# Assert noise is positive
		noise = jnp.array(noise)
		noise = eqx.error_if(noise, jnp.any(noise < 0), "noise must be positive or nul.")

		# Set the computation engine to SafeDiagonalEngine for diagonal computation
		kwargs['computation_engine'] = SafeDiagonalEngine
		super().__init__(**kwargs)

		# Transform to unconstrained space
		self._raw_noise = to_unconstrained(jnp.asarray(noise))




import equinox as eqx
import jax.numpy as jnp
from equinox import filter_jit
from jax import Array

from ..AbstractKernel import AbstractKernel, StaticAbstractKernel
from ..utils import format_jax_array


class StaticConstantKernel(StaticAbstractKernel):
	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute the constant kernel covariance value.

		:param kern: the kernel to use, containing hyperparameters
		:param x1: scalar array (ignored)
		:param x2: scalar array (ignored)
		:return: scalar array (constant value)
		"""
		kern = eqx.combine(kern)
		return kern.value  # type: ignore[no-any-return,attr-defined]


class ConstantKernel(AbstractKernel):
	"""
	Constant Kernel

	Returns a constant value regardless of inputs.
	The value parameter can be any real number (positive or negative).
	No transformation is applied since there's no positivity constraint.
	"""

	value: Array = eqx.field(converter=jnp.asarray)
	static_class = StaticConstantKernel

	def __init__(self, value=1.0, **kwargs):
		"""
		Initialize the Constant kernel.

		Args:
			value: The constant value to return (can be any real number)
		"""
		super().__init__(**kwargs)
		self.value = jnp.asarray(value)

	def __str__(self):
		return format_jax_array(self.value)



from jax import numpy as jnp
from jax import Array
import equinox as eqx
from equinox import filter_jit

from ..AbstractKernel import AbstractKernel
from ..distances import squared_euclidean_distance
from ..transforms import to_constrained, to_unconstrained
from .StationaryKernel import StaticStationaryKernel


class StaticFeatureKernel(StaticStationaryKernel):
	""" Static class allowing FeatureKernel to be used inside a BlockKernel. """

	distance_func = squared_euclidean_distance

	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern: AbstractKernel, x1: jnp.ndarray, x2: jnp.ndarray) -> jnp.ndarray:
		"""
		Compute the kernel covariance value between two vectors.

		:param kern: kernel instance containing the hyperparameters
		:param x1: scalar array
		:param x2: scalar array
		:return: scalar array
		"""
		kern = eqx.combine(kern)

		# If length_scale or variance is a scalar, convert it to 2-element array for consistency
		if kern.length_scale.ndim == 0:
			kern_length_scale = jnp.array([kern.length_scale, kern.length_scale])
		else:
			kern_length_scale = kern.length_scale
		if kern.variance.ndim == 0:
			kern_variance = jnp.array([kern.variance, kern.variance])
		else:
			kern_variance = kern.variance

		# As the formula only involves diagonal matrices, we can compute directly with vectors
		sigma_diag = kern_length_scale[0] + kern_length_scale[1] + kern.length_scale_u  # Σ
		sigma_det = jnp.prod(sigma_diag)  # |Σ|

		# Compute the quadratic form: (x - x')^T Sigma^{-1} (x - x')
		# Since Sigma^{-1} is diagonal, this simplifies to sum of (diff_i^2 * sigma_inv_diag_i)
		quadratic_form = cls.distance_func(x1, x2) / sigma_diag

		return kern_variance[0] * kern_variance[1] / (((2 * jnp.pi)**(len(x1)/2)) * jnp.sqrt(sigma_det)) * jnp.exp(-0.5 * quadratic_form)


class FeatureKernel(AbstractKernel):
	"""
	Feature Kernel with multiple positive-constrained length scales and variances.

	All parameters (length_scale_1, length_scale_2, length_scale_u, variance_1, variance_2)
	are constrained to be positive.
	"""

	_raw_length_scale: Array = eqx.field(converter=jnp.asarray)
	_raw_length_scale_u: Array = eqx.field(converter=jnp.asarray)
	_raw_variance: Array = eqx.field(converter=jnp.asarray)

	static_class = StaticFeatureKernel

	def __init__(self, length_scale, length_scale_u, variance, **kwargs):
		"""
		Initialize the Feature kernel.

		:param length_scale: length scale parameter - must be positive
		:param length_scale_u: uncertainty length scale parameter - must be positive
		:param variance: variance parameter - must be positive
		"""
		# Validate all parameters are positive
		length_scale = jnp.array(length_scale)
		length_scale_u = jnp.array(length_scale_u)
		variance = jnp.array(variance)

		length_scale = eqx.error_if(length_scale, jnp.any(length_scale <= 0), "length_scale must be positive.")
		length_scale_u = eqx.error_if(length_scale_u, jnp.any(length_scale_u <= 0), "length_scale_u must be positive.")
		variance = eqx.error_if(variance, jnp.any(variance <= 0), "variance must be positive.")

		# Initialize parent
		super().__init__(**kwargs)

		# Store in unconstrained space
		self._raw_length_scale = to_unconstrained(jnp.asarray(length_scale))
		self._raw_length_scale_u = to_unconstrained(jnp.asarray(length_scale_u))
		self._raw_variance = to_unconstrained(jnp.asarray(variance))

	@property
	def length_scale(self) -> Array:
		"""Get length_scale_1 in constrained (positive) space."""
		return to_constrained(self._raw_length_scale)

	@property
	def length_scale_u(self) -> Array:
		"""Get length_scale_u in constrained (positive) space."""
		return to_constrained(self._raw_length_scale_u)

	@property
	def variance(self) -> Array:
		"""Get variance_1 in constrained (positive) space."""
		return to_constrained(self._raw_variance)



"""
Classes that extend Equinox' Module to provide state-structure separation.

A module is either a Static class or an instance. The instance is linked to a static class
to make computations.

These classes are mainly here to be extended by Kernel and Mean classes, while providing
type-checking for functions that receive a Module as an argument.
"""

from typing import TYPE_CHECKING, ClassVar, Optional, Type
import equinox as eqx
import jax.numpy as jnp
from jax import Array

if TYPE_CHECKING:
	pass


class StaticAbstractModule:
	pass  # Nothing special here


class AbstractModule(eqx.Module):
	static_class: ClassVar[Optional[Type[StaticAbstractModule]]] = None

	def __init__(self, **kwargs):
		"""
		Initialize the module and mark that a module has been instantiated.

		This locks the parameter_transform config setting to prevent inconsistencies
		with JIT-compiled code.

		Args:
			**kwargs: Additional keyword arguments (for subclass compatibility)
		"""
		super().__init__(**kwargs)

		# Import here to avoid circular dependency
		from .config import config

		# Mark that kernels have been instantiated (locks parameter_transform)
		config._mark_module_instantiated()

	def __call__(self, *args, **kwargs):
		raise NotImplementedError("Subclasses must implement __call__ method.")

	def replace(self, **kwargs):
		"""API de modification fonctionnelle (Setter idiomatique)."""
		from .transforms import to_unconstrained

		adapted = {}
		for k, v in kwargs.items():
			raw_field = f"_raw_{k}"

			if hasattr(self, raw_field):
				v = jnp.asarray(v)
				v = eqx.error_if(v, jnp.any(v <= 0), f"{k} must be positive.")

				current = getattr(self, raw_field)
				if current.shape != v.shape:
					v = jnp.broadcast_to(v, current.shape)

				adapted[raw_field] = to_unconstrained(v)
			elif hasattr(self, k):
				current = getattr(self, k)

				if isinstance(current, Array):
					v = jnp.asarray(v) if not isinstance(v, Array) else v

					if current.shape != v.shape:
						v = jnp.broadcast_to(v, current.shape)

				adapted[k] = v

		where = lambda s: [getattr(s, k) for k in adapted.keys()]
		return eqx.tree_at(where, self, list(adapted.values()))

	def __add__(self, other):
		from kernax.operators import SumModule

		return SumModule(self, other)

	def __radd__(self, other):
		from kernax.operators import SumModule

		return SumModule(other, self)

	def __sub__(self, other):
		from kernax.operators import SumModule
		from kernax.wrappers import NegModule

		return SumModule(self, NegModule(other))

	def __rsub__(self, other):
		from kernax.operators import SumModule
		from kernax.wrappers import NegModule

		return SumModule(other, NegModule(self))

	def __neg__(self):
		from kernax.wrappers import NegModule

		return NegModule(self)

	def __mul__(self, other):
		from kernax.operators import ProductModule

		return ProductModule(self, other)

	def __rmul__(self, other):
		from kernax.operators import ProductModule

		return ProductModule(other, self)

	def __str__(self):
		from kernax.transforms import to_constrained
		from kernax.utils import format_jax_array

		return f"{self.__class__.__name__}({
			', '.join(
				[
					f'{key}={format_jax_array(value)}' if '_raw_' not in key else f'{key[5:]}={format_jax_array(to_constrained(value))}'
					for key, value in self.__dict__.items()
					if isinstance(value, Array)
				]
				+
				[
					f'{key}={value}' for key, value in self.__dict__.items() if isinstance(value, (int, float, str))
				]
			)
		})"


from __future__ import annotations

from typing import ClassVar, Optional, Type

import jax.numpy as jnp
from equinox import filter_jit
from jax import Array, vmap

from .module import AbstractModule, StaticAbstractModule


class AbstractMean(AbstractModule):
	static_class: ClassVar[Optional[Type[StaticAbstractMean]]] = None

	def __init__(self, **kwargs):
		super().__init__(**kwargs)

	@filter_jit
	def __call__(self, x: Array) -> Array:
		x = jnp.atleast_1d(x)

		assert self.static_class is not None, "static_class must be defined in subclass"

		if jnp.ndim(x) == 1:
			return self.static_class.scalar_mean(self, x)
		elif jnp.ndim(x) == 2:
			return self.static_class.vector_mean(self, x)
		else:
			raise ValueError(
				f"Invalid input dimensions: x has shape {x.shape}. "
				"Expected 1D or 2D arrays as inputs."
			)


class StaticAbstractMean(StaticAbstractModule):
	@classmethod
	@filter_jit
	def scalar_mean(cls, mean: AbstractMean, x: Array) -> Array:
		"""
		Compute the mean value for a single input vector.

		:param mean: mean instance containing the parameters
		:param x: 1D array
		:return: scalar array
		"""
		return jnp.array(jnp.nan)  # To be overwritten in subclasses

	@classmethod
	@filter_jit
	def vector_mean(cls, mean: AbstractMean, x: Array) -> Array:
		"""
		Compute the mean for a batch of input vectors.

		:param mean: mean instance containing the parameters
		:param x: 2D array (N, D)
		:return: 1D array (N,)
		"""
		return vmap(cls.scalar_mean, in_axes=(None, 0))(mean, x)


import jax.scipy as jsp
from equinox import filter_jit
from jax import Array

from .BatchModule import BatchModule


class BlockDiagKernel(BatchModule):
	"""
	Wrapper kernel to build block-diagonal covariance matrices using any kernel.

	A basic kernel usually works on inputs of shape (N, I), and produces covariance matrices of shape (N, N).

	Wrapped inside a block-diagonal kernel, they can either:
	- still work on inputs of shape (N, I), but produce covariance matrices of shape (B*N, B*N), where B is the number of blocks. This is useful when the hyperparameters are distinct to blocks, i.e. each sub-matrix has its own set of hyperparameters.
	- or work on inputs of shape (B, N, I), producing covariance matrices of shape (B*N, B*N). This is useful when inputs are different for each block, regardless of whether the hyperparameters are shared between blocks or not.

	This class uses vmap to vectorize the kernel computation of each block, then resize the result into a block matrix.
	"""

	def __init__(self, inner, nb_blocks, block_in_axes=None, block_over_inputs=True, **kwargs):
		super().__init__(inner, nb_blocks, block_in_axes, block_over_inputs, **kwargs)

	@filter_jit
	def __call__(self, x1: Array, x2: None | Array = None) -> Array:
		"""
		Compute the kernel over batched inputs using vmap.

		Args:
				x1: Input of shape (B, ..., N, I)
				x2: Optional second input of shape (B, ..., M, I)

		Returns:
				Kernel block-matrix of appropriate shape
		"""
		return jsp.linalg.block_diag(*super().__call__(x1, x2))  # type: ignore[no-any-return]

	def __str__(self):
		return f"BlockDiag{self.inner}"



from __future__ import annotations

from typing import Optional

from equinox import filter_jit
from jax import Array

from .WrapperModule import WrapperModule


class NegModule(WrapperModule):
	@filter_jit
	def __call__(self, x1: Array, x2: Optional[Array] = None) -> Array:
		if x2 is None:
			return -self.inner(x1)
		return -self.inner(x1, x2)

	def __str__(self):
		return f"- {self.inner}"


"""Dot-product based kernels."""

from .LinearKernel import LinearKernel, StaticLinearKernel
from .PolynomialKernel import PolynomialKernel, StaticPolynomialKernel
from .Sigmoid import SigmoidKernel, StaticSigmoidKernel

__all__ = [
	"LinearKernel",
	"StaticLinearKernel",
	"PolynomialKernel",
	"StaticPolynomialKernel",
	"SigmoidKernel",
	"StaticSigmoidKernel",
]



import equinox as eqx
import jax.numpy as jnp
import jax.tree_util as jtu
from equinox import filter_jit
from jax import Array

from .WrapperModule import WrapperModule


class ARDKernel(WrapperModule):
	"""
	Wrapper kernel to apply Automatic Relevance Determination (ARD) to the inputs before passing them to the inner kernel.
	Each input dimension is scaled by a separate length scale hyperparameter.
	"""

	length_scales: Array = eqx.field(converter=jnp.asarray)

	def _freeze_inner_lengthscales(self, kernel):
		def map_func(path, leaf):
			if len(path) > 0:
				last_node = path[-1]  # To retrieve the attribute name
				if isinstance(last_node, jtu.GetAttrKey) and last_node.name == "length_scale":
					# Force length scale of 1
					return jnp.ones_like(leaf)
			return leaf

		return jtu.tree_map_with_path(map_func, kernel)

	def __init__(self, inner, length_scales, **kwargs):
		"""
		:param inner: the kernel to wrap, must be an instance of AbstractKernel
		:param length_scales: the length scales for each input dimension (1D array of floats)
		"""
		super().__init__(inner=inner, **kwargs)
		self.length_scales = length_scales

	@filter_jit
	def __call__(self, x1: Array, x2: None | Array = None) -> Array:
		# TODO: add runtime error if length_scales doesn't match input dimensions
		if x2 is None:
			x2 = x1

		# FIXME: if used in an optimisation setting, inner length_scales can still have other values.
		#  Freezing the inner at every call may be costly/slow down learning.
		#  We should find a proper way to freeze these parameters.
		return self._freeze_inner_lengthscales(self.inner)(  # type: ignore[no-any-return]
			x1 / self.length_scales, x2 / self.length_scales
		)



import equinox as eqx
from equinox import filter_jit
from jax import Array
from jax import numpy as jnp

from ..AbstractKernel import AbstractKernel
from .DotProductKernel import StaticDotProductKernel


class StaticPolynomialKernel(StaticDotProductKernel):
	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern: AbstractKernel, x1: Array, x2: Array) -> Array:
		"""
		Compute the polynomial kernel covariance value between two vectors.

		:param kern: kernel instance containing the hyperparameters
		:param x1: scalar array
		:param x2: scalar array
		:return: scalar array
		"""
		kern = eqx.combine(kern)
		dp = cls.distance_func(x1, x2)
		return jnp.pow(kern.gamma * dp + kern.constant, kern.degree)  # type: ignore[attr-defined]


class PolynomialKernel(AbstractKernel):
	"""
	Polynomial Kernel

	Parameter gamma is always positive.
	Parameter constant can be any real value.
	Parameter degree is a static integer.
	"""

	degree: int = eqx.field(static=True)
	_raw_gamma: Array = eqx.field(converter=jnp.asarray)
	constant: Array = eqx.field(converter=jnp.asarray)

	static_class = StaticPolynomialKernel

	def __init__(self, degree: int, gamma: float = 1.0, constant: float = 0.0, **kwargs):
		"""
		Initialize the Polynomial kernel.

		Args:
			degree: Degree of the polynomial (must be positive integer)
			gamma: Scale factor (must be positive)
			constant: Independent term (can be any real value)

		Raises:
			ValueError: If degree is not positive or gamma is not positive
		"""
		# Validate degree
		if degree <= 0:
			raise ValueError(f"degree must be a positive integer, got {degree}")

		# Validate gamma positivity
		gamma_array = jnp.array(gamma)
		gamma_array = eqx.error_if(gamma_array, jnp.any(gamma_array <= 0), "gamma must be positive.")

		# Initialize parent (locks config)
		super().__init__(**kwargs)

		# Set static degree
		self.degree = degree

		# Transform gamma to unconstrained space
		from ..transforms import to_unconstrained

		self._raw_gamma = to_unconstrained(jnp.asarray(gamma_array))

		# constant can be any value, no transformation needed
		self.constant = jnp.asarray(constant)

	@property
	def gamma(self) -> Array:
		"""Get the gamma parameter in constrained space (always positive)."""
		from ..transforms import to_constrained

		return to_constrained(self._raw_gamma)



"""
kernax-ml: A JAX-based kernel library for Gaussian Processes.

kernax-ml provides a collection of kernel functions (covariance functions) for
Gaussian Process models, with support for automatic differentiation, JIT
compilation, and composable kernel operations.
"""

__version__ = "0.5.0-alpha"
__author__ = "S. Lejoly"
__email__ = "simon.lejoly@unamur.be"
__license__ = "MIT"

# Import transformation utilities
from . import transforms
from .module import AbstractModule, StaticAbstractModule
from .AbstractKernel import AbstractKernel, StaticAbstractKernel
from .AbstractMean import AbstractMean, StaticAbstractMean

# Import configuration system
from .config import config

# Import dot-product kernels
from .dotproduct import (
	LinearKernel,
	PolynomialKernel,
	SigmoidKernel,
	StaticLinearKernel,
	StaticPolynomialKernel,
	StaticSigmoidKernel,
)

# Import operator modules
from .operators import (
	OperatorModule,
	ProductModule,
	SumModule,
)

# Import other kernels
from .other import (
	ConstantKernel,
	StaticConstantKernel,
	VarianceKernel,
	WhiteNoiseKernel,
)

# Import stationary kernels
from .stationary import (
	Matern12Kernel,
	Matern32Kernel,
	Matern52Kernel,
	PeriodicKernel,
	RationalQuadraticKernel,
	RBFKernel,
	SEKernel,
	FeatureKernel,
	StaticMatern12Kernel,
	StaticMatern32Kernel,
	StaticMatern52Kernel,
	StaticPeriodicKernel,
	StaticRationalQuadraticKernel,
	StaticSEKernel,
	StaticFeatureKernel,
)

# Import wrapper modules/kernels
from .wrappers import (
	ActiveDimsModule,
	ARDKernel,
	BatchModule,
	BlockDiagKernel,
	BlockKernel,
	ExpModule,
	LogModule,
	NegModule,
	WrapperModule,
)

# Import mean functions
from .means import (
	AffineMean,
	ConstantMean,
	LinearMean,
	StaticAffineMean,
	StaticConstantMean,
	StaticLinearMean,
	StaticZeroMean,
	ZeroMean,
)

__all__ = [
	# Package metadata
	"__version__",
	"__author__",
	"__email__",
	"__license__",
	# Configuration
	"config",
	# Transformations
	"transforms",
	# Base classes
	"StaticAbstractModule",
	"AbstractModule",
	"StaticAbstractKernel",
	"AbstractKernel",
	"StaticAbstractMean",
	"AbstractMean",
	# Base kernels
	"StaticSEKernel",
	"SEKernel",
	"RBFKernel",
	"StaticConstantKernel",
	"ConstantKernel",
	"StaticLinearKernel",
	"LinearKernel",
	"StaticPeriodicKernel",
	"PeriodicKernel",
	"StaticRationalQuadraticKernel",
	"RationalQuadraticKernel",
	"StaticPolynomialKernel",
	"PolynomialKernel",
	"StaticSigmoidKernel",
	"SigmoidKernel",
	"VarianceKernel",
	"WhiteNoiseKernel",
	"StaticFeatureKernel",
	"FeatureKernel",
	# Matern family
	"StaticMatern12Kernel",
	"Matern12Kernel",
	"StaticMatern32Kernel",
	"Matern32Kernel",
	"StaticMatern52Kernel",
	"Matern52Kernel",
	# Operator modules
	"OperatorModule",
	"SumModule",
	"ProductModule",
	# Wrapper modules/kernels
	"WrapperModule",
	"NegModule",
	"ExpModule",
	"LogModule",
	"ActiveDimsModule",
	"ARDKernel",
	"BatchModule",
	"BlockKernel",
	"BlockDiagKernel",
	# Mean functions
	"StaticZeroMean",
	"ZeroMean",
	"StaticConstantMean",
	"ConstantMean",
	"StaticLinearMean",
	"LinearMean",
	"StaticAffineMean",
	"AffineMean",
]


"""Other kernel types."""

from .ConstantKernel import ConstantKernel, StaticConstantKernel
from .VarianceKernel import VarianceKernel
from .WhiteNoiseKernel import WhiteNoiseKernel

__all__ = [
	"ConstantKernel",
	"StaticConstantKernel",
	"VarianceKernel",
	"WhiteNoiseKernel",
]



from __future__ import annotations

import equinox as eqx
import jax.numpy as jnp
from equinox import filter_jit
from jax import Array

from ..AbstractMean import AbstractMean, StaticAbstractMean


class StaticLinearMean(StaticAbstractMean):
	@classmethod
	@filter_jit
	def scalar_mean(cls, mean: AbstractMean, x: Array) -> Array:
		return jnp.sum(mean.slope * x)


class LinearMean(AbstractMean):
	slope: Array = eqx.field(converter=jnp.asarray)
	static_class = StaticLinearMean

	def __init__(self, slope=0.0, **kwargs):
		super().__init__(**kwargs)
		self.slope = jnp.asarray(slope)


"""Stationary kernels."""

from .Matern12Kernel import Matern12Kernel, StaticMatern12Kernel
from .Matern32Kernel import Matern32Kernel, StaticMatern32Kernel
from .Matern52Kernel import Matern52Kernel, StaticMatern52Kernel
from .PeriodicKernel import PeriodicKernel, StaticPeriodicKernel
from .RationalQuadraticKernel import RationalQuadraticKernel, StaticRationalQuadraticKernel
from .SEKernel import RBFKernel, SEKernel, StaticSEKernel
from .FeatureKernel import FeatureKernel, StaticFeatureKernel

__all__ = [
	"SEKernel",
	"StaticSEKernel",
	"RBFKernel",
	"PeriodicKernel",
	"StaticPeriodicKernel",
	"RationalQuadraticKernel",
	"StaticRationalQuadraticKernel",
	"Matern12Kernel",
	"StaticMatern12Kernel",
	"Matern32Kernel",
	"StaticMatern32Kernel",
	"Matern52Kernel",
	"StaticMatern52Kernel",
	"FeatureKernel",
	"StaticFeatureKernel",
]



from .AffineMean import AffineMean, StaticAffineMean
from .ConstantMean import ConstantMean, StaticConstantMean
from .LinearMean import LinearMean, StaticLinearMean
from .ZeroMean import StaticZeroMean, ZeroMean

__all__ = [
	"StaticZeroMean",
	"ZeroMean",
	"StaticConstantMean",
	"ConstantMean",
	"StaticLinearMean",
	"LinearMean",
	"StaticAffineMean",
	"AffineMean",
]


from .ActiveDimsModule import ActiveDimsModule
from .ARDKernel import ARDKernel
from .BatchModule import BatchModule
from .BlockDiagKernel import BlockDiagKernel
from .BlockKernel import BlockKernel
from .ExpModule import ExpModule
from .LogModule import LogModule
from .NegModule import NegModule
from .WrapperModule import WrapperModule

__all__ = [
	"WrapperModule",
	"ExpModule",
	"LogModule",
	"NegModule",
	"ActiveDimsModule",
	"BatchModule",
	"BlockKernel",
	"BlockDiagKernel",
	"ARDKernel",
]



from __future__ import annotations

import jax.numpy as jnp
from equinox import filter_jit
from jax import Array

from ..AbstractMean import AbstractMean, StaticAbstractMean


class StaticZeroMean(StaticAbstractMean):
	@classmethod
	@filter_jit
	def scalar_mean(cls, mean: AbstractMean, x: Array) -> Array:
		return jnp.array(0.0)


class ZeroMean(AbstractMean):
	static_class = StaticZeroMean